﻿using System;
using Ozeki.Media;
using Ozeki.Media.MediaHandlers;
using Ozeki.MediaGateway;
using Ozeki.MediaGateway.Service;
using Ozeki.Network;
using Ozeki.Network.Nat;
using Ozeki.VoIP;
using Ozeki.VoIP.SDK;

namespace Web2SIPServer
{
    class MyClient
    {

        // Client reference
        IClient client;

        ISoftPhone softPhone;
        IPhoneLine phoneLine;
        IPhoneCall call;

        IStreamService streamService;
        string publishStreamName;
        string playStreamName;

        MediaConnector mediaConnector;
        PhoneCallAudioSender phoneCallAudioSender;
        PhoneCallAudioReceiver phoneCallAudioReceiver;

        MediaGatewayAudioReceiver mediaGatewayAudioReceiver;
        MediaGatewayAudioSender mediaGatewayAudioSender;

        /// <summary>
        /// Creates a new instance of MyClient.
        /// </summary>
        /// <param name="client">The connected client.</param>
        /// <param name="streamService"> Stream service what transmits media data between two client.</param>
        /// <param name="softPhone"> Represents a SIP softphone</param>
        /// <param name="account">SIP account for PBX registration.</param>
        public MyClient(IClient client, IStreamService streamService, ISoftPhone softPhone, SIPAccount account)
        {
            mediaConnector = new MediaConnector();
            phoneCallAudioSender = new PhoneCallAudioSender();
            phoneCallAudioReceiver = new PhoneCallAudioReceiver();

            this.streamService = streamService;
            this.softPhone = softPhone;
            this.client = client;

            Account = account;
        }

        /// <summary>
        /// SIP account for able to register a SIP PBX.
        /// </summary>
        public SIPAccount Account { get; private set; }


        /// <summary>
        /// Registers a SIP account for a SIP PBX.
        /// </summary>
        public virtual void Register()
        {
            phoneLine = softPhone.CreatePhoneLine(Account, new NatConfiguration(NatTraversalMethod.None), TransportType.Udp);

            phoneLine.PhoneLineStateChanged += PhoneLine_PhoneLineInformationCanged;
            softPhone.IncomingCall += SoftPhone_IncomingCall;


            softPhone.RegisterPhoneLine(phoneLine);

        }

        /// <summary>
        /// Unregisters the registered phoneline from the SIP PBX. 
        /// </summary>
        public virtual void Unregister()
        {
            if (mediaGatewayAudioReceiver != null)
            {
                streamService.RemoveStream(mediaGatewayAudioReceiver.MediaStream);
                mediaGatewayAudioReceiver = null;
            }

            HangUp();

            softPhone.IncomingCall -= SoftPhone_IncomingCall;

            if (phoneLine != null)
            {
                phoneLine.PhoneLineStateChanged -= PhoneLine_PhoneLineInformationCanged;
                softPhone.UnregisterPhoneLine(phoneLine);
            }
        }


        /// <summary>
        /// Call the given dial number.
        /// </summary>
        /// <param name="dialNumber">dial number</param>
        public virtual void Call(string dialNumber)
        {
            if (phoneLine == null)
            {
                System.Diagnostics.Debug.Fail("PhoneLine null");
                return;
            }

            if ((phoneLine.RegisteredInfo == PhoneLineState.RegistrationSucceeded || phoneLine.RegisteredInfo == PhoneLineState.NoRegNeeded) && call == null)
            {
                call = softPhone.CreateCallObject(phoneLine, dialNumber);

                call.CallErrorOccured += Call_ErrorOccured;
                call.CallStateChanged += Call_StateChanged;
                call.Start();
            }
        }

        /// <summary>
        /// Hangup the call.
        /// </summary>
        public virtual void HangUp()
        {
            if (call != null)
            {
                call.HangUp();
                call = null;
            }
        }

        /// <summary>
        /// Accept the incoming call request.
        /// </summary>
        public virtual void Accept()
        {
            if (call != null)
            {
                call.Accept();
            }
        }
        /// <summary>
        /// Reject the incoming call request.
        /// </summary>
        public void Reject()
        {
            if (call != null)
            {
                call.Reject();
                call = null;
            }
        }

        public void StreamPublished(IMediaStream mediaStream)
        {
            mediaGatewayAudioSender = new MediaGatewayAudioSender(mediaStream);
            mediaConnector.Connect(mediaGatewayAudioSender, phoneCallAudioSender);
        }

        public void StreamClosed()
        {
            if (mediaGatewayAudioSender != null)
            {
                mediaConnector.Disconnect(mediaGatewayAudioSender, phoneCallAudioSender);
                mediaGatewayAudioSender.Dispose();
                mediaGatewayAudioSender = null;
            }

        }

        #region VoIP SDK events

        /// <summary>
        /// Indicates phoneline information changes.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void PhoneLine_PhoneLineInformationCanged(object sender, VoIPEventArgs<PhoneLineState> e)
        {
            if (e.Item == PhoneLineState.RegistrationSucceeded || e.Item == PhoneLineState.NoRegNeeded)
                OnSetReadyStatus(true, Account.DisplayName);
            else
            {
                OnSetReadyStatus(false, e.Item.ToString());
            }
        }

        /// <summary>
        /// Indicates received incoming call.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void SoftPhone_IncomingCall(object sender, VoIPEventArgs<IPhoneCall> e)
        {

            if (e.Item.PhoneLine == phoneLine)
            {
                call = e.Item;
                call.CallErrorOccured += Call_ErrorOccured;
                call.CallStateChanged += Call_StateChanged;

                OnCallRequest(call.DialInfo.DisplayName);
            }
        }

        /// <summary>
        /// Indicates call state changes.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        protected virtual void Call_StateChanged(object sender, VoIPEventArgs<CallState> e)
        {
            switch (e.Item)
            {
                case CallState.InCall:
                    phoneCallAudioSender.AttachToCall(call);
                    phoneCallAudioReceiver.AttachToCall(call);

                    publishStreamName = Guid.NewGuid().ToString(); // Ez az amit a kliens publikál
                    playStreamName = Guid.NewGuid().ToString(); // Ez az amit a szerver publikál

                    var mediaStream = streamService.CreateStream(playStreamName);
                    mediaGatewayAudioReceiver = new MediaGatewayAudioReceiver(mediaStream);

                    mediaConnector.Connect(phoneCallAudioReceiver, mediaGatewayAudioReceiver);

                    OnInCall(publishStreamName, playStreamName);
                    break;

                case CallState.Completed:

                    phoneCallAudioSender.Detach();
                    phoneCallAudioReceiver.Detach();
                    
                    streamService.RemoveStream(mediaGatewayAudioReceiver.MediaStream);
                    mediaConnector.Disconnect(phoneCallAudioReceiver, mediaGatewayAudioReceiver);
                    mediaGatewayAudioReceiver.Dispose();

                    mediaGatewayAudioReceiver = null;
                    call = null;
                    break;
            }

            OnCallStateChanged(e.Item);
        }

        /// <summary>
        /// Indicates call errors. For example : The called number can not be found, equal with Callstate.NotFound
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        protected virtual void Call_ErrorOccured(object sender, VoIPEventArgs<CallError> e)
        {
            OnCallErrorOccurred(e.Item);
        }

        #endregion

        #region This methods invoke client side methods.

        /// <summary>
        /// Indicates for the client, his call state has changed.
        /// </summary>
        /// <param name="callState"></param>
        void OnCallStateChanged(CallState callState)
        {
            Console.WriteLine("OnCallStateChanged");
            client.InvokeMethod("OnCallStateChanged", callState);
        }
        /// <summary>
        /// Indicates for the client, his call request ended with error.
        /// </summary>
        /// <param name="callError"></param>
        void OnCallErrorOccurred(CallError callError)
        {
            Console.WriteLine("OnCallErrorOccurred");
            client.InvokeMethod("OnCallErrorOccurred", callError);
        }


    
        /// <summary>
        /// Invokes client's callback method, to indicates for the client he can ready to start a call or receive a request.
        /// </summary>
        /// <param name="isReady">ready signal.</param>
        /// <param name="name"> The client name.</param>
        public void OnSetReadyStatus(bool isReady, string name)
        {
            try
            {
                client.InvokeMethod("OnSetReadyStatus", isReady, name);
            }
            catch (Exception)
            { }
        }

        /// <summary>
        /// Invokes client's callback method, to indicates for the client he has got a new call request from other client.
        /// </summary>
        /// <param name="requestOwner">The call request sender's clientId.</param>
        public void OnCallRequest(string requestOwner)
        {
            Console.WriteLine("Call request received from {0} to {1}", requestOwner, "Name");
            
          
            client.InvokeMethod("OnCallRequest", requestOwner);
        }

        /// <summary>
        /// Jel küldése a kliensnek hozzon létre és egy MediaStreamSender és publikálja az első paraméterrel, valamint egy MediaStreamReceivert és játsza le a második paramétert
        /// </summary>
        /// <param name="senderPublishID">A kliens MediaStreamSender-ének ezen a néven kell publikálnia a stream-ét. </param>
        /// <param name="receiverPlayID"> A kliens MediaStreamReceiver-ének ezt a nevű szerver által létrehozott streamet kell lejátszani.</param>
        public void OnInCall(string senderPublishID, string  receiverPlayID)
        {
            Console.WriteLine("Sends 'start publishing' sign to the clients.");
            client.InvokeMethod("OnInCall",senderPublishID,receiverPlayID);
         
        }

        #endregion
    }
}
